# mypy: ignore-errors
import json
from typing import Any, Dict, List, Optional, Callable, cast, Union

from openai import AsyncOpenAI, BadRequestError, RateLimitError, APIError, AuthenticationError

from .base import BaseAgent, AgentResponse, AgentToolCall
from .logger import get_logger
from .permissions import PermissionOptions, PermissionRequest, PermissionStatus
from .tools.register_tools import register_default_tools

# Initialize logger
logger = get_logger(__name__)


class OpenAIAgent(BaseAgent):
    """
    OpenAI Agent that implements the BaseAgent interface using OpenAI's models.
    """

    def __init__(
        self,
        api_key: str,
        model: str = "gpt-4-turbo",
        temperature: float = 0.0,
        timeout: int = 180,
        permission_callback: Optional[Callable[[PermissionRequest], PermissionStatus]] = None,
        permission_options: Optional[PermissionOptions] = None,
        default_tool_timeout: int = 300,
        **kwargs
    ):
        """
        Initialize an OpenAI agent.

        Args:
            api_key: OpenAI API key
            model: OpenAI model to use, default is gpt-4-turbo
            temperature: Temperature parameter for model (0.0 to 1.0)
            timeout: Timeout in seconds for API requests
            permission_callback: Optional callback for permission requests
            permission_options: Permission configuration options
            default_tool_timeout: Default timeout in seconds for tool calls (default: 300s)
            **kwargs: Additional parameters to pass to the model
        """
        logger.info(f"Initializing OpenAI agent with model {model}")

        super().__init__(
            api_key=api_key,
            model=model,
            permission_options=permission_options,
            permission_callback=permission_callback,
            default_tool_timeout=default_tool_timeout
        )

        self.temperature = temperature
        self.timeout = timeout
        self.extra_kwargs = kwargs

        # Initialize OpenAI client
        self.client = AsyncOpenAI(api_key=api_key)
        logger.debug("Initialized OpenAI client")

        self.conversation_history = []
        self.available_tools = {}
        self.system_prompt = self._generate_system_prompt()
        logger.debug(f"Generated system prompt ({len(self.system_prompt)} chars)")
        logger.debug(f"Tool timeouts set to {default_tool_timeout}s")

    def _is_valid_api_key(self, api_key: str) -> bool:
        """
        Validate the format of the OpenAI API key.

        Args:
            api_key: The API key to validate

        Returns:
            True if the key is a valid format, False otherwise
        """
        if not api_key or not isinstance(api_key, str):
            logger.warning("API key is empty or not a string")
            return False

        # OpenAI keys should start with sk-
        valid_prefix = api_key.startswith("sk-")

        # Keys should be fairly long and not contain spaces
        valid_length = len(api_key) >= 20 and " " not in api_key

        if not valid_prefix or not valid_length:
            logger.warning("Invalid API key format")

        return valid_prefix and valid_length

    def _generate_system_prompt(self) -> str:
        """
        Generate the system prompt that defines the agent's capabilities and behavior.
        This is an extensive prompt that replicates Claude's behavior in Cursor.
        """
        logger.debug("Generating system prompt for OpenAI agent")
        return """
You are a powerful agentic AI coding assistant, powered by OpenAI's advanced models. You operate exclusively in Cursor, the world's best IDE.

You are pair programming with a USER to solve their coding task.
The task may require creating a new codebase, modifying or debugging an existing codebase, or simply answering a question.
Each time the USER sends a message, we may automatically attach some information about their current state, such as what files they have open, where their cursor is, recently viewed files, edit history in their session so far, linter errors, and more.
This information may or may not be relevant to the coding task, it is up for you to decide.
Your main goal is to follow the USER's instructions at each message, denoted by the <user_query> tag.

<tool_calling>
You have tools at your disposal to solve the coding task. Follow these rules regarding tool calls:
1. ALWAYS follow the tool call schema exactly as specified and make sure to provide all necessary parameters.
2. The conversation may reference tools that are no longer available. NEVER call tools that are not explicitly provided.
3. **NEVER refer to tool names when speaking to the USER.** For example, instead of saying 'I need to use the edit_file tool to edit your file', just say 'I will edit your file'.
4. Only calls tools when they are necessary. If the USER's task is general or you already know the answer, just respond without calling tools.
5. Before calling each tool, first explain to the USER why you are calling it.
</tool_calling>

<making_code_changes>
When making code changes, NEVER output code to the USER, unless requested. Instead use one of the code edit tools to implement the change.
Use the code edit tools at most once per turn.
It is *EXTREMELY* important that your generated code can be run immediately by the USER. To ensure this, follow these instructions carefully:
1. Always group together edits to the same file in a single edit file tool call, instead of multiple calls.
2. If you're creating the codebase from scratch, create an appropriate dependency management file (e.g. requirements.txt) with package versions and a helpful README.
3. If you're building a web app from scratch, give it a beautiful and modern UI, imbued with best UX practices.
4. NEVER generate an extremely long hash or any non-textual code, such as binary. These are not helpful to the USER and are very expensive.
5. Unless you are appending some small easy to apply edit to a file, or creating a new file, you MUST read the the contents or section of what you're editing before editing it.
6. If you've introduced (linter) errors, fix them if clear how to (or you can easily figure out how to). Do not make uneducated guesses. And DO NOT loop more than 3 times on fixing linter errors on the same file. On the third time, you should stop and ask the user what to do next.
7. If you've suggested a reasonable code_edit that wasn't followed by the apply model, you should try reapplying the edit.
</making_code_changes>

<searching_and_reading>
You have tools to search the codebase and read files. Follow these rules regarding tool calls:
1. If available, heavily prefer the semantic search tool to grep search, file search, and list dir tools.
2. If you need to read a file, prefer to read larger sections of the file at once over multiple smaller calls.
3. If you have found a reasonable place to edit or answer, do not continue calling tools. Edit or answer from the information you have found.
</searching_and_reading>

Answer the user's request using the relevant tool(s), if they are available. Check that all the required parameters for each tool call are provided or can reasonably be inferred from context. IF there are no relevant tools or there are missing values for required parameters, ask the user to supply these values; otherwise proceed with the tool calls. If the user provides a specific value for a parameter (for example provided in quotes), make sure to use that value EXACTLY. DO NOT make up values for or ask about optional parameters. Carefully analyze descriptive terms in the request as they may indicate required parameter values that should be included even if not explicitly quoted.

You MUST use the following format when citing code regions or blocks:
```12:15:app/components/Todo.tsx
// ... existing code ...
```
This is the ONLY acceptable format for code citations. The format is ```startLine:endLine:filepath where startLine and endLine are line numbers.
"""

    def _prepare_tools(self) -> Optional[List[Dict[str, Any]]]:
        """
        Prepare the registered tools for OpenAI API.

        Returns:
            List of tools in the format expected by OpenAI API, or None if no tools are registered
        """
        if not self.available_tools:
            logger.debug("No tools registered")
            return None

        logger.debug(f"Preparing {len(self.available_tools)} tools for OpenAI API")
        tools = []
        for name, tool_data in self.available_tools.items():
            tools.append(
                {
                    "type": "function",
                    "function": {
                        "name": name,
                        "description": tool_data["schema"]["description"],
                        "parameters": {
                            "type": "object",
                            "properties": tool_data["schema"]["parameters"]["properties"],
                            "required": tool_data["schema"]["parameters"].get("required", []),
                        },
                    },
                }
            )
            logger.debug(f"Prepared tool: {name}")

        return tools if tools else None

    def _execute_tool_calls(self, tool_calls: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Execute the tool calls made by OpenAI.

        Args:
            tool_calls: List of tool calls to execute

        Returns:
            List of tool call results
        """
        logger.info(f"Executing {len(tool_calls)} tool calls")
        tool_results = []

        for call in tool_calls:
            try:
                # Handle both dict format and ChatCompletionMessageToolCall objects
                if hasattr(call, "function"):
                    # It's an object
                    tool_name = call.function.name
                    try:
                        arguments = json.loads(call.function.arguments)
                    except json.JSONDecodeError:
                        arguments = {}
                    tool_call_id = call.id
                    logger.debug(f"Executing tool (object): {tool_name} (id: {tool_call_id})")
                else:
                    # It's a dict
                    tool_name = call["function"]["name"]
                    try:
                        arguments = json.loads(call["function"]["arguments"])
                    except json.JSONDecodeError:
                        arguments = {}
                    # Cast to str to handle potential missing 'id' attribute
                    tool_call_id = cast(str, call.get("id", "unknown_id"))
                    logger.debug(f"Executing tool (dict): {tool_name} (id: {tool_call_id})")

                logger.debug(f"Tool arguments: {json.dumps(arguments)}")

                if tool_name not in self.available_tools:
                    logger.warning(f"Tool not found: {tool_name}")
                    result = {
                        "role": "tool",
                        "tool_call_id": tool_call_id,
                        "content": f"Error: Tool '{tool_name}' not found",
                    }
                else:
                    logger.debug(f"Calling function for tool: {tool_name}")
                    function = self.available_tools[tool_name]["function"]
                    result_content = function(**arguments)

                    # Log a summary of the result
                    if isinstance(result_content, dict) and "error" in result_content:
                        logger.warning(f"Tool {tool_name} returned error: {result_content.get('error')}")
                    else:
                        content_preview = str(result_content)
                        if len(content_preview) > 100:
                            content_preview = content_preview[:100] + "..."
                        logger.debug(f"Tool {tool_name} result: {content_preview}")

                    result = {
                        "role": "tool",
                        "tool_call_id": tool_call_id,
                        "content": (
                            json.dumps(result_content)
                            if isinstance(result_content, dict)
                            else str(result_content)
                        ),
                    }
                tool_results.append(result)
            except Exception as e:
                # Log the error - in production, this would go to a proper logging system
                logger.error(f"Error executing tool call: {str(e)}")
                # We still need to add a response to maintain the conversation flow
                if "tool_call_id" in locals():
                    tool_results.append(
                        {
                            "role": "tool",
                            "tool_call_id": tool_call_id,  # type: ignore
                            "content": f"Error: {str(e)}",
                        }
                    )

        logger.info(f"Completed {len(tool_results)} tool call results")
        return tool_results

    async def chat(self, message: str, user_info: Optional[Dict[str, Any]] = None) -> Union[str, AgentResponse]:
        """
        Send a message to the OpenAI API and get a response.

        Args:
            message: The user's message
            user_info: Optional dict containing info about the user's current state

        Returns:
            Either a string response (for backward compatibility) or a structured AgentResponse
            containing the message, tool_calls made, and optional thinking
        """
        # Format the user message with user_info if provided
        formatted_message = self.format_user_message(message, user_info)

        logger.info("Sending message to OpenAI API")
        logger.debug(f"Message length: {len(formatted_message)} chars")

        # Add the user message to the conversation history
        self.conversation_history.append({"role": "user", "content": formatted_message})

        # Prepare the messages for the API call
        messages = [{"role": "system", "content": self.system_prompt}] + self.conversation_history
        logger.debug(f"Total context: {len(messages)} messages")

        # Prepare tools
        tools = self._prepare_tools()

        # Initialize the structured response
        processed_tool_calls: List[AgentToolCall] = []

        try:
            # Make the API call
            logger.debug(f"Calling OpenAI API with model: {self.model or 'gpt-4-turbo'}")
            if tools:
                logger.debug(f"Using {len(tools)} tools")

            response = await self.client.chat.completions.create(  # type: ignore
                model=self.model if self.model else "gpt-4-turbo",
                messages=messages,
                tools=tools,
                tool_choice="auto" if tools else None,
                max_tokens=4096,
                temperature=self.temperature,
            )
            logger.info("Received response from OpenAI API")

            # Get the assistant's response
            assistant_message = response.choices[0].message

            # Track thinking (not directly supported by OpenAI but we can add it in the future)
            thinking = None

            # Check if there are any tool calls
            if assistant_message.tool_calls:
                logger.info(f"Response contains {len(assistant_message.tool_calls)} tool calls")

                # Add the assistant's response to the conversation history
                self.conversation_history.append(
                    {
                        "role": "assistant",
                        "content": assistant_message.content or "",
                        "tool_calls": assistant_message.tool_calls,
                    }
                )

                # Execute the tool calls
                tool_results = self._execute_tool_calls(assistant_message.tool_calls)

                # Process and track tool calls for the structured response
                for idx, tool_call in enumerate(assistant_message.tool_calls):
                    tool_name = tool_call.function.name
                    try:
                        parameters = json.loads(tool_call.function.arguments)
                    except json.JSONDecodeError:
                        parameters = {}

                    # Find the corresponding result
                    result = None
                    for res in tool_results:
                        if res.get("tool_call_id") == tool_call.id:
                            result = res.get("content", "")
                            break

                    # Add to processed tool calls
                    processed_tool_calls.append({
                        "name": tool_name,
                        "parameters": parameters,
                        "result": result
                    })

                # Add the tool results to the conversation history
                for result in tool_results:
                    self.conversation_history.append(result)

                # Make a follow-up API call with the tool results
                logger.debug("Making follow-up API call with tool results")
                follow_up_messages = (
                    messages
                    + [
                        {
                            "role": "assistant",
                            "content": assistant_message.content or "",
                            "tool_calls": [
                                {
                                    "id": tool_call.id,
                                    "function": {
                                        "name": tool_call.function.name,
                                        "arguments": tool_call.function.arguments,
                                    },
                                    "type": "function",
                                }
                                for tool_call in assistant_message.tool_calls
                            ],
                        }
                    ]
                    + tool_results
                )
                logger.debug(f"Follow-up call with {len(follow_up_messages)} messages")

                follow_up_response = await self.client.chat.completions.create(
                    model=self.model if self.model else "gpt-4-turbo", messages=follow_up_messages, max_tokens=4096, temperature=self.temperature
                )
                logger.info("Received follow-up response from OpenAI API")

                # Add the assistant's follow-up response to the conversation history
                follow_up_message = follow_up_response.choices[0].message
                self.conversation_history.append(
                    {"role": "assistant", "content": follow_up_message.content}
                )

                response_text = follow_up_message.content or ""
                logger.debug(f"Follow-up response text length: {len(response_text)} chars")

                # Return structured response
                return {
                    "message": response_text,
                    "tool_calls": processed_tool_calls,
                    "thinking": thinking
                }
            else:
                # Add the assistant's response to the conversation history
                self.conversation_history.append(
                    {"role": "assistant", "content": assistant_message.content}
                )

                response_text = assistant_message.content or ""
                logger.debug(f"Response text length: {len(response_text)} chars")

                # Return structured response
                return {
                    "message": response_text,
                    "tool_calls": processed_tool_calls,
                    "thinking": thinking
                }

        except AuthenticationError as e:
            error_msg = f"Error: Authentication failed. Please check your OpenAI API key. Details: {str(e)}"
            logger.error(f"Authentication error: {str(e)}")
            return {
                "message": error_msg,
                "tool_calls": [],
                "thinking": None
            }
        except BadRequestError as e:
            error_msg = f"Error: Bad request to the OpenAI API. Details: {str(e)}"
            logger.error(f"Bad request error: {str(e)}")
            return {
                "message": error_msg,
                "tool_calls": [],
                "thinking": None
            }
        except RateLimitError as e:
            error_msg = f"Error: Rate limit exceeded. Please try again later. Details: {str(e)}"
            logger.error(f"Rate limit error: {str(e)}")
            return {
                "message": error_msg,
                "tool_calls": [],
                "thinking": None
            }
        except APIError as e:
            error_msg = f"Error: OpenAI API error. Details: {str(e)}"
            logger.error(f"API error: {str(e)}")
            return {
                "message": error_msg,
                "tool_calls": [],
                "thinking": None
            }
        except Exception as e:
            error_msg = f"Error: An unexpected error occurred. Details: {str(e)}"
            logger.error(f"Unexpected error: {type(e).__name__}: {str(e)}")
            return {
                "message": error_msg,
                "tool_calls": [],
                "thinking": None
            }

    def register_default_tools(self) -> None:
        """
        Register all the default tools available to the agent.
        """
        # Use the centralized tool registration function
        logger.info("Registering default tools")
        register_default_tools(self)

    def _permission_request_callback(self, permission_request: PermissionRequest) -> PermissionStatus:
        """
        Implementation of permission request callback for OpenAI agent.

        In a real application, this would interact with the user to get permission.
        For now, we'll default to a console-based interaction.

        Args:
            permission_request: The permission request object

        Returns:
            PermissionStatus indicating whether the request is granted or denied
        """
        # If yolo mode is enabled, check is already done in PermissionManager
        if self.permission_manager.options.yolo_mode:
            logger.debug(f"Permission automatically granted in yolo mode for {permission_request.operation}")
            return PermissionStatus.GRANTED

        # Default implementation asks on console
        logger.debug(f"Requesting permission for {permission_request.operation}")
        print(f"\n[PERMISSION REQUEST] {permission_request.operation}")
        print(f"Details: {json.dumps(permission_request.details, indent=2)}")
        response = input("Allow this operation? (y/n): ").strip().lower()

        if response == 'y' or response == 'yes':
            logger.debug("Permission granted by user")
            return PermissionStatus.GRANTED
        else:
            logger.debug("Permission denied by user")
            return PermissionStatus.DENIED

    async def query_image(self, image_paths: List[str], query: str) -> str:
        """
        Query the OpenAI model about one or more images.

        Args:
            image_paths: List of paths to local image files
            query: The query/question about the image(s)

        Returns:
            The model's response about the image(s)
        """
        import os
        import base64

        logger.info(f"Processing image query with {len(image_paths)} images")

        # Validate image paths
        for path in image_paths:
            if not os.path.exists(path):
                error_msg = f"Image file not found: {path}"
                logger.error(error_msg)
                return error_msg

        # Prepare images for the API
        image_contents = []
        for path in image_paths:
            try:
                with open(path, "rb") as image_file:
                    # Get file extension without the dot
                    file_extension = os.path.splitext(path)[1][1:].lower()
                    # Map file extension to MIME type
                    mime_types = {
                        "jpg": "image/jpeg",
                        "jpeg": "image/jpeg",
                        "png": "image/png",
                        "gif": "image/gif",
                        "webp": "image/webp"
                    }
                    mime_type = mime_types.get(file_extension, "application/octet-stream")

                    # Encode image as base64
                    encoded_image = base64.b64encode(image_file.read()).decode('utf-8')

                    image_contents.append({
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:{mime_type};base64,{encoded_image}"
                        }
                    })
                    logger.debug(f"Processed image: {path} ({mime_type})")
            except Exception as e:
                error_msg = f"Error processing image {path}: {str(e)}"
                logger.error(error_msg)
                return error_msg

        try:
            # Prepare the message with images and query
            messages = [
                {"role": "system", "content": "You are an AI assistant that can analyze images."},
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": query},
                        *image_contents
                    ]
                }
            ]

            # Call the OpenAI API with GPT-4 Vision
            logger.debug(f"Calling OpenAI API for image analysis with model: {self.model}")
            vision_model = "gpt-4o" if self.model.startswith("gpt-4") else "gpt-4o"

            response = await self.client.chat.completions.create(
                model=vision_model,
                messages=messages,
                max_tokens=1024,
                temperature=self.temperature,
                timeout=self.timeout
            )

            # Extract and return the assistant's response
            if response.choices and len(response.choices) > 0:
                result = response.choices[0].message.content or ""
                logger.info("Successfully processed image query")
                return result
            else:
                error_msg = "No response received from OpenAI API"
                logger.error(error_msg)
                return error_msg

        except BadRequestError as e:
            error_msg = f"Bad request to OpenAI API: {str(e)}"
            logger.error(error_msg)
            return error_msg
        except RateLimitError as e:
            error_msg = f"Rate limit exceeded: {str(e)}"
            logger.error(error_msg)
            return error_msg
        except APIError as e:
            error_msg = f"OpenAI API error: {str(e)}"
            logger.error(error_msg)
            return error_msg
        except Exception as e:
            error_msg = f"Unexpected error processing image query: {str(e)}"
            logger.error(error_msg)
            return error_msg
